import logging
from abc import ABC, abstractmethod
from typing import Dict, List

from kombu import Connection, Exchange, Message, Queue
from kombu.mixins import ConsumerProducerMixin
from pycapo import CapoConfig

"""
Class Messenger: for creating the Messenger object.
The messenger object is responsible for creating AMQP connections, exchanges,
and queues for a workspace service.
The messenger object is responsible for sending and receiving 
messages from AMQP directly.
The messenger object is also responsible for message-to-event
 and event-to-message translation
"""


logger = logging.getLogger()


class AMQPServerInitializer:
    """
    Object responsible for creating a connection to the correct AMQP server and establishing exchanges and queues for
    that server
    """

    def __init__(self, service_name: str):
        self.service_name = service_name
        self.connection = self._setup_connection()

    @staticmethod
    def _setup_connection() -> Connection:
        """
        Setup connection to the Workspaces AMQP server

        :return: AMQP connection
        """
        config = CapoConfig().settings("edu.nrao.archive.configuration.AmqpServer")

        return Connection(
            hostname=config.hostname,
            port=int(config.port),
            userid=config.username,
            password=config.password,
            connect_timeout=int(5),
        )

    @staticmethod
    def create_exchange() -> Exchange:
        """
        Create AMQP topic exchange given a name

        :return: Newly created exchange
        """
        return Exchange("workspaces", type="topic")

    def create_queues(self, exchange: Exchange) -> List[Queue]:
        """
        Create correct queues for exchange given a service name and set them up for use
        TODO: Refactor to only create a "persistent queue for the user"

        :param exchange: Name of the AMQP exchange
        :return: List of newly created and set-up queues
        """
        q_name = self.service_name + "_queue"

        if self.service_name == "capability":
            cp_queue = Queue(q_name, exchange, routing_key=self.service_name)
            cp_wf_queue = Queue("cp_wf_queue", exchange, routing_key="workflow")
            queues = [cp_queue, cp_wf_queue]
        elif self.service_name == "workflow":
            queues = [Queue(q_name, exchange, routing_key=self.service_name)]
        else:
            return []

        self.setup_queues(queues)
        return queues

    def setup_queues(self, queues: List[Queue]):
        """
        Perform setup for AMQP queues

        :param queues: List of queues to set up
        """
        for queue in queues:
            queue.maybe_bind(self.connection)
            queue.declare()


class MessageReceiverIF(ABC):
    """
    Abstract interface for a message-receiving object
    """

    @abstractmethod
    def receive_message(self, message):
        pass


class Messenger(ConsumerProducerMixin):
    """
    Object responsible for:
    1. Maintaining a connection to the Workspaces AMQP server
    2. Consuming messages from the server and transfering them to a destination for further processing
    3. Sending messages to the server
    """

    def __init__(self, destination: MessageReceiverIF, service_name: str):
        server_initializer = AMQPServerInitializer(service_name)
        self.connection = server_initializer.connection
        self.exchange = server_initializer.create_exchange()
        self.queues = server_initializer.create_queues(self.exchange)
        self.destination = destination

    def receive_message(self, message: Message):
        """
        Callback invoked upon consuming of an AMQP message; sends consumed message to a MessageReceiver and
        acknowledges message's receipt

        :param message: AMQP message
        """
        try:
            self.destination.receive_message(message.body)
        except Exception as e:
            # NOTE: If we want to develop retrying message handling, this is probably where it should happen
            logger.exception("exception in callback handling for message %s", message.body)
        finally:
            message.ack()

    def get_consumers(self, Consumer, channel):
        """
        Implementation of abstract method from superclass ConsumerProducerMixin; returns a list of queues that this
        object will consume from
        """
        consumers = []
        for queue in self.queues:
            consumers.append(
                Consumer(
                    queues=queue,
                    on_message=self.receive_message,
                    prefetch_count=10,
                )
            )
        logger.info(self.queues)
        return consumers

    def send_message(self, **message: Dict):
        """
        Send message to AMQP server

        :param message: Body of AMQP message
        """
        if "routing_key" not in message.keys():
            logger.warning("Message has no routing key.")

        self.producer.publish(
            message,
            routing_key=message["routing_key"],
            exchange=self.exchange,
            # TODO: Figure out routing key format
            retry=True,
        )
